#!/usr/bin/python3

# Copyright (c) 2020 ModalAI Inc.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# 4. The Software is used solely in conjunction with devices provided by
#    ModalAI Inc.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# For a license to use on non-ModalAI hardware, please contact license@modalai.com

from voxl_esc_setup_paths import *
voxl_esc_setup_paths()

from libesc import *
from libesc.esctypes import EscTypes as types
from esc_scanner import EscScanner
from esc_boards import *
import time
import argparse

parser = argparse.ArgumentParser(description='ESC Upload Parameters Script')
parser.add_argument('-d','--device',         required=False, default=None)
parser.add_argument('-b','--baud-rate',      required=False, default=None)
parser.add_argument('-i','--id',             type=int,  required=False,  default=255)
parser.add_argument('-p','--params-file',    type=str, required=True,  default="")
parser.add_argument('-f','--params-filter',  type=str, required=False, default="all")
parser.add_argument('-n','--esc-count',      type=int, required=False, default=0)
parser.add_argument('-g','--enable-debug',   required=False, default=0)
parser.add_argument('-v','--verify-params',  required=False, default=0)
args = parser.parse_args()

devpath       = args.device
baudrate      = args.baud_rate
params_file   = args.params_file
params_filter = args.params_filter
esc_count     = args.esc_count
esc_id        = args.id
debug         = False
verify_params = bool(args.verify_params)

try:
    debug=args.enable_debug == 1 or 'True' in args.enable_debug or 'true' in args.enable_debug or '1' in args.enable_debug
except Exception as e:
    pass

compatible_firmware_versions = [38, 39]


print('\nINFO: Params file name : %s ' %(params_file))
params_file_size = 0
try:
    params_file_size = os.path.getsize(params_file)
    print('INFO: Params file size : %d bytes' % (params_file_size))
except:
    raise
    print('ERROR: Unable to open params file %s' % (params_file))
    sys.exit(1)
print('\n')

esc_ids_found = [0,0,0,0]

if baudrate is None:
    firmware_baudrates   = [2000000, 250000, 921600, 230400, 57600, 115200]
else:
    firmware_baudrates = [baudrate]

for b in firmware_baudrates:
    baudrate_test = b
    devpath_test  = args.device

    # quick scan for ESCs to detect the port
    if 1: #devpath is None or baudrate is None:
        scanner = EscScanner()
        if debug:
            scanner.enable_debug()
        (devpath, baudrate) = scanner.scan(devpath_test, baudrate_test)

    if devpath is not None and baudrate is not None:
        print('INFO: ESC(s) detected on port: ' + devpath + ', baud rate: ' + str(baudrate))
    else:
        print('ERROR: No ESC(s) detected')
        continue
        #sys.exit(1)

    # create ESC manager and search for ESCs
    try:
        esc_manager = EscManager()
        if debug:
            esc_manager.enable_debug()
        esc_manager.open(devpath, baudrate)
    except Exception as e:
        print('ERROR: Unable to connect to ESCs :')
        print(e)
        sys.exit(1)

    #time.sleep(0.25)
    escs = esc_manager.get_escs()
    esc_ids = [e.get_id() for e in escs]

    print('INFO: ESCs detected:')
    print('INFO: ---------------------')

    for e in esc_manager.get_escs():
        esc_ids_found[e.get_id()] = 1
        versions = e.get_versions()
        hardware_name = get_esc_board_description(versions[1])
        print('ID: %d, SW: %d, HW: %d: %s' % (e.get_id(), versions[0], versions[1], hardware_name))
    print('---------------------')

    #if (esc_count != 0 and len(escs) != esc_count) or (len(escs) == 0):
    #if len(escs) == 0:
    if 0:
        print('ERROR: Only %d ESCs detected, exiting.' % len(escs))
        #print('ERROR: No ESCs detected, exiting.')
        esc_manager.close()
        sys.exit(1)

    #check firmware version for param compatibility
    for e in esc_manager.get_escs():
        versions = e.get_versions()
        if versions[0] not in compatible_firmware_versions:
            print('ERROR: ESC firmware version %d is incompatible with params supported by this utility' % (versions[0]))
            print('       Acceptable ESC firmware versions are: ', compatible_firmware_versions)
            sys.exit(1)

    esc = esc_manager.esc_dummy

    # check for .xml or .eep (binary) config files
    filename, file_extension = os.path.splitext(params_file)
    if os.path.isfile(params_file) and file_extension == '.xml':
        print('INFO: Loading XML config file...')
        with open(params_file, 'r') as file:
            xml_string = file.read()
            esc.params.parse_xml_string( xml_string )

    #elif os.path.isfile(params_file) and file_extension == '.eep':
    #    print('INFO: Loading EEP config file...')
    #    with open(params_file, 'rb') as file:
    #        param_bytes = file.read()
    #        esc.params.parse_params_all( param_bytes )
    else:
        print('ERROR: Unsupported params file type, exiting. : %s' % (params_file))
        esc_manager.close()
        sys.exit(1)

    #additional param checking
    param_baud_rate  = esc.params.get_param_val(types.ESC_PACKET_TYPE_UART_CONFIG, "baud_rate")
    valid_baud_rates = [57600, 115200, 230400, 250000, 460800, 921600, 1000000, 2000000]
    if param_baud_rate not in valid_baud_rates:
        print('ERROR: Invalid param baud rate : %d' % (param_baud_rate))
        print('Valid baud rates: ',valid_baud_rates)
        esc_manager.close()
        sys.exit(1)

    if esc_id == 255:
        print('INFO: Uploading params for all ESC IDs ...')
    else:
        print('INFO: Uploading params for ESC ID %d...' % esc_id)

    if 'all' in params_filter or 'board' in params_filter:
        print('-- board config')
        esc_manager.push_config_data(types.ESC_PACKET_TYPE_BOARD_CONFIG, esc_id)
        time.sleep(0.1)

    if 'all' in params_filter or 'id' in params_filter:
        print('-- id config')
        esc_manager.push_config_data(types.ESC_PACKET_TYPE_ID_CONFIG, esc_id)
        time.sleep(0.1)

    if 'all' in params_filter or 'uart' in params_filter:
        print('-- uart config')
        esc_manager.push_config_data(types.ESC_PACKET_TYPE_UART_CONFIG, esc_id)
        time.sleep(0.1)

    if 'all' in params_filter or 'tune' in params_filter:
        print('-- tune config')
        esc_manager.push_config_data(types.ESC_PACKET_TYPE_TUNE_CONFIG, esc_id)
        time.sleep(0.1)

    print('    DONE')
    
    # reset all ESCs
    #time.sleep(0.5)
    print('INFO: Resetting ESCs...')
    if (esc_id == 255):
        esc_manager.reset_all()
    else:
        esc_manager.reset_id(esc_id);
    print('    DONE')
    esc_manager.close()

    # sleep before reset just in case
    time.sleep(1)

    if all(esc_ids_found):
        break


if verify_params == True:
    print('Waiting for ESCs to restart..')
    time.sleep(2)
    verify_params_script = './voxl-esc-verify-params.py'
    print('Verifying Params.. Calling %s' % verify_params_script)
    import subprocess
    subprocess.call(verify_params_script, shell=True)

